@@ -8,20 +8,17 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 """
-This module contains definitions and functions for XSD builtin datatypes.
+This module contains definitions and a factory function for XSD builtin datatypes.
 
 Only atomic builtins are created, the list builtins types ('NMTOKENS', 'ENTITIES', 'IDREFS')
 are created using the XSD 1.0 meta-schema or with and additional base schema for XSD 1.1.
 """
-import re
-import base64
 from decimal import Decimal
-from math import isinf, isnan
-
 from elementpath import datatypes
+from typing import Any, Dict, Optional, Type, Tuple, Union
 
 from ..exceptions import XMLSchemaValueError
-from ..qnames import XSD_LENGTH, XSD_MIN_LENGTH, XSD_MAX_LENGTH, XSD_ENUMERATION, \
+from ..names import XSD_LENGTH, XSD_MIN_LENGTH, XSD_MAX_LENGTH, XSD_ENUMERATION, \
     XSD_PATTERN, XSD_WHITE_SPACE, XSD_MIN_INCLUSIVE, XSD_MIN_EXCLUSIVE, XSD_MAX_INCLUSIVE, \
     XSD_MAX_EXCLUSIVE, XSD_TOTAL_DIGITS, XSD_FRACTION_DIGITS, XSD_EXPLICIT_TIMEZONE, \
     XSD_STRING, XSD_NORMALIZED_STRING, XSD_NAME, XSD_NCNAME, XSD_QNAME, XSD_TOKEN, \
@@ -35,160 +32,42 @@ from ..qnames import XSD_LENGTH, XSD_MIN_LENGTH, XSD_MAX_LENGTH, XSD_ENUMERATION
     XSD_HEX_BINARY, XSD_NOTATION_TYPE, XSD_ERROR, XSD_ASSERTION, XSD_SIMPLE_TYPE, \
     XSD_ANY_TYPE, XSD_ANY_ATOMIC_TYPE, XSD_ANY_SIMPLE_TYPE
 from ..etree import etree_element
-from ..helpers import is_etree_element
-from .exceptions import XMLSchemaValidationError
+from ..aliases import ElementType, SchemaType, BaseXsdType
+
+from .helpers import decimal_validator, qname_validator, byte_validator, \
+    short_validator, int_validator, long_validator, unsigned_byte_validator, \
+    unsigned_short_validator, unsigned_int_validator, unsigned_long_validator, \
+    negative_int_validator, positive_int_validator, non_positive_int_validator, \
+    non_negative_int_validator, hex_binary_validator, base64_binary_validator, \
+    error_type_validator, boolean_to_python, python_to_boolean
 from .facets import XSD_10_FACETS_BUILDERS, XSD_11_FACETS_BUILDERS
 from .simple_types import XsdSimpleType, XsdAtomicBuiltin
 
-HEX_BINARY_PATTERN = re.compile(r'^[0-9a-fA-F]+$')
-BASE64_BINARY_PATTERN = re.compile(
-    r'((([A-Za-z0-9+/] ?){4})*(([A-Za-z0-9+/] ?){3}[A-Za-z0-9+/]|([A-Za-z0-9+/] ?){2}'
-    r'[AEIMQUYcgkosw048] ?=|[A-Za-z0-9+/] ?[AQgw] ?= ?=))?'
-)
-
 #
 # Admitted facets sets for XSD atomic types
-STRING_FACETS = (
+STRING_FACETS = {
     XSD_LENGTH, XSD_MIN_LENGTH, XSD_MAX_LENGTH, XSD_PATTERN,
     XSD_ENUMERATION, XSD_WHITE_SPACE, XSD_ASSERTION
-)
+}
 
-BOOLEAN_FACETS = (XSD_PATTERN, XSD_WHITE_SPACE, XSD_ASSERTION)
+BOOLEAN_FACETS = {XSD_PATTERN, XSD_WHITE_SPACE, XSD_ASSERTION}
 
-FLOAT_FACETS = (
+FLOAT_FACETS = {
     XSD_PATTERN, XSD_ENUMERATION, XSD_WHITE_SPACE, XSD_MAX_INCLUSIVE,
     XSD_MAX_EXCLUSIVE, XSD_MIN_INCLUSIVE, XSD_MIN_EXCLUSIVE, XSD_ASSERTION
-)
+}
 
-DECIMAL_FACETS = (
+DECIMAL_FACETS = {
     XSD_TOTAL_DIGITS, XSD_FRACTION_DIGITS, XSD_PATTERN, XSD_ENUMERATION,
     XSD_WHITE_SPACE, XSD_MAX_INCLUSIVE, XSD_MAX_EXCLUSIVE, XSD_MIN_INCLUSIVE,
     XSD_MIN_EXCLUSIVE, XSD_ASSERTION
-)
+}
 
-DATETIME_FACETS = (
+DATETIME_FACETS = {
     XSD_PATTERN, XSD_ENUMERATION, XSD_WHITE_SPACE,
     XSD_MAX_INCLUSIVE, XSD_MAX_EXCLUSIVE, XSD_MIN_INCLUSIVE,
     XSD_MIN_EXCLUSIVE, XSD_ASSERTION, XSD_EXPLICIT_TIMEZONE
-)
-
-
-#
-# XSD built-in types validator functions
-def decimal_validator(x):
-    try:
-        if isinf(x) or isnan(x) or 'E' in str(x).upper():
-            yield XMLSchemaValidationError(decimal_validator, x,
-                                           "value {!r} is not a valid xs:decimal".format(x))
-    except TypeError:
-        pass
-
-
-def qname_validator(x):
-    if datatypes.QNAME_PATTERN.match(x) is None:
-        yield XMLSchemaValidationError(qname_validator, x,
-                                       "value {!r} is not an xs:QName".format(x))
-
-
-def byte_validator(x):
-    if not (-2**7 <= x < 2**7):
-        yield XMLSchemaValidationError(int_validator, x, "value must be -128 <= x < 128.")
-
-
-def short_validator(x):
-    if not (-2**15 <= x < 2**15):
-        yield XMLSchemaValidationError(short_validator, x, "value must be -2^15 <= x < 2^15.")
-
-
-def int_validator(x):
-    if not (-2**31 <= x < 2**31):
-        yield XMLSchemaValidationError(int_validator, x, "value must be -2^31 <= x < 2^31.")
-
-
-def long_validator(x):
-    if not (-2**63 <= x < 2**63):
-        yield XMLSchemaValidationError(long_validator, x, "value must be -2^63 <= x < 2^63.")
-
-
-def unsigned_byte_validator(x):
-    if not (0 <= x < 2**8):
-        yield XMLSchemaValidationError(unsigned_byte_validator, x, "value must be 0 <= x < 256.")
-
-
-def unsigned_short_validator(x):
-    if not (0 <= x < 2**16):
-        yield XMLSchemaValidationError(unsigned_short_validator, x, "value must be 0 <= x < 2^16.")
-
-
-def unsigned_int_validator(x):
-    if not (0 <= x < 2**32):
-        yield XMLSchemaValidationError(unsigned_int_validator, x, "value must be 0 <= x < 2^32.")
-
-
-def unsigned_long_validator(x):
-    if not (0 <= x < 2**64):
-        yield XMLSchemaValidationError(unsigned_long_validator, x, "value must be 0 <= x < 2^64.")
-
-
-def negative_int_validator(x):
-    if x >= 0:
-        yield XMLSchemaValidationError(negative_int_validator, x, reason="value must be negative.")
-
-
-def positive_int_validator(x):
-    if x <= 0:
-        yield XMLSchemaValidationError(positive_int_validator, x, "value must be positive.")
-
-
-def non_positive_int_validator(x):
-    if x > 0:
-        yield XMLSchemaValidationError(non_positive_int_validator, x, "value must be non positive.")
-
-
-def non_negative_int_validator(x):
-    if x < 0:
-        yield XMLSchemaValidationError(non_negative_int_validator, x, "value must be non negative.")
-
-
-def hex_binary_validator(x):
-    if x and (len(x) % 2 or HEX_BINARY_PATTERN.match(x) is None):
-        yield XMLSchemaValidationError(hex_binary_validator, x, "not an hexadecimal number.")
-
-
-def base64_binary_validator(value):
-    value = value.replace(' ', '')
-    if not value:
-        return
-
-    match = BASE64_BINARY_PATTERN.match(value)
-    if match is None or match.group(0) != value:
-        yield XMLSchemaValidationError(base64_binary_validator, value, "not a base64 encoding")
-    else:
-        try:
-            base64.standard_b64decode(value)
-        except (ValueError, TypeError) as err:
-            yield XMLSchemaValidationError(base64_binary_validator, value,
-                                           "not a base64 encoding: %s." % err)
-
-
-def error_type_validator(x):
-    yield XMLSchemaValidationError(error_type_validator, x,
-                                   "not value is allowed for xs:error type.")
-
-
-#
-# XSD builtin decoding functions
-def boolean_to_python(s):
-    if s in ('true', '1'):
-        return True
-    elif s in ('false', '0'):
-        return False
-    else:
-        raise XMLSchemaValueError('not a boolean value: %r' % s)
-
-
-def python_to_boolean(obj):
-    return str(obj).lower()
+}
 
 
 #
@@ -222,41 +101,42 @@ XSD_COMMON_BUILTIN_TYPES = (
         'name': XSD_DECIMAL,
         'python_type': (Decimal, str, int, float),
         'admitted_facets': DECIMAL_FACETS,
+        'to_python': datatypes.DecimalProxy,
         'facets': [decimal_validator, COLLAPSE_WHITE_SPACE_ELEMENT],
     },  # decimal number
 
     # --- Dates and Times (not year related) ---
     {
         'name': XSD_GDAY,
-        'python_type': (str, datatypes.GregorianDay),
+        'python_type': (datatypes.GregorianDay, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianDay.fromstring,
     },  # DD
     {
         'name': XSD_GMONTH,
-        'python_type': (str, datatypes.GregorianMonth),
+        'python_type': (datatypes.GregorianMonth, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianMonth.fromstring,
     },  # MM
     {
         'name': XSD_GMONTH_DAY,
-        'python_type': (str, datatypes.GregorianMonthDay),
+        'python_type': (datatypes.GregorianMonthDay, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianMonthDay.fromstring,
     },  # MM-DD
     {
         'name': XSD_TIME,
-        'python_type': (str, datatypes.Time),
+        'python_type': (datatypes.Time, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.Time.fromstring,
     },  # hh:mm:ss
     {
         'name': XSD_DURATION,
-        'python_type': (str, datatypes.Duration),
+        'python_type': (datatypes.Duration, str),
         'admitted_facets': FLOAT_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.Duration.fromstring,
@@ -291,13 +171,13 @@ XSD_COMMON_BUILTIN_TYPES = (
     },  # true/false or 1/0
     {
         'name': XSD_BASE64_BINARY,
-        'python_type': str,
+        'python_type': (datatypes.Base64Binary, str, bytes),
         'admitted_facets': STRING_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT, base64_binary_validator],
     },  # base64 encoded binary value
     {
         'name': XSD_HEX_BINARY,
-        'python_type': str,
+        'python_type': (datatypes.HexBinary, str, bytes),
         'admitted_facets': STRING_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT, hex_binary_validator],
     },   # hexadecimal encoded binary value
@@ -323,9 +203,7 @@ XSD_COMMON_BUILTIN_TYPES = (
         'name': XSD_LANGUAGE,
         'python_type': str,
         'base_type': XSD_TOKEN,
-        'facets': [
-            etree_element(XSD_PATTERN, value=r"[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*")
-        ]
+        'facets': [etree_element(XSD_PATTERN, value=r"[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*")]
     },  # language codes
     {
         'name': XSD_NAME,
@@ -450,7 +328,7 @@ XSD_COMMON_BUILTIN_TYPES = (
     },  # only negative value allowed [< 0]
 )
 
-XSD_10_BUILTIN_TYPES = XSD_COMMON_BUILTIN_TYPES + (
+XSD_10_BUILTIN_TYPES: Tuple[Dict[str, Any], ...] = XSD_COMMON_BUILTIN_TYPES + (
     {
         'name': XSD_DOUBLE,
         'python_type': float,
@@ -467,35 +345,35 @@ XSD_10_BUILTIN_TYPES = XSD_COMMON_BUILTIN_TYPES + (
     # --- Year related primitive types (year 0 not allowed) ---
     {
         'name': XSD_DATETIME,
-        'python_type': (str, datatypes.DateTime10),
+        'python_type': (datatypes.DateTime10, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.DateTime10.fromstring,
     },  # [-][Y*]YYYY-MM-DD[Thh:mm:ss]
     {
         'name': XSD_DATE,
-        'python_type': (str, datatypes.Date10),
+        'python_type': (datatypes.Date10, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.Date10.fromstring,
     },  # [-][Y*]YYYY-MM-DD
     {
         'name': XSD_GYEAR,
-        'python_type': (str, datatypes.GregorianYear10),
+        'python_type': (datatypes.GregorianYear10, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianYear10.fromstring,
     },  # [-][Y*]YYYY
     {
         'name': XSD_GYEAR_MONTH,
-        'python_type': (str, datatypes.GregorianYearMonth10),
+        'python_type': (datatypes.GregorianYearMonth10, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianYearMonth10.fromstring,
     },  # [-][Y*]YYYY-MM
 )
 
-XSD_11_BUILTIN_TYPES = XSD_COMMON_BUILTIN_TYPES + (
+XSD_11_BUILTIN_TYPES: Tuple[Dict[str, Any], ...] = XSD_COMMON_BUILTIN_TYPES + (
     {
         'name': XSD_DOUBLE,
         'python_type': float,
@@ -512,28 +390,28 @@ XSD_11_BUILTIN_TYPES = XSD_COMMON_BUILTIN_TYPES + (
     # --- Year related primitive types (year 0 allowed and mapped to 1 BCE) ---
     {
         'name': XSD_DATETIME,
-        'python_type': (str, datatypes.DateTime),
+        'python_type': (datatypes.DateTime, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.DateTime.fromstring,
     },  # [-][Y*]YYYY-MM-DD[Thh:mm:ss]
     {
         'name': XSD_DATE,
-        'python_type': (str, datatypes.Date),
+        'python_type': (datatypes.Date, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.Date.fromstring,
     },  # [-][Y*]YYYY-MM-DD
     {
         'name': XSD_GYEAR,
-        'python_type': (str, datatypes.GregorianYear),
+        'python_type': (datatypes.GregorianYear, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianYear.fromstring,
     },  # [-][Y*]YYYY
     {
         'name': XSD_GYEAR_MONTH,
-        'python_type': (str, datatypes.GregorianYearMonth),
+        'python_type': (datatypes.GregorianYearMonth, str),
         'admitted_facets': DATETIME_FACETS,
         'facets': [COLLAPSE_WHITE_SPACE_ELEMENT],
         'to_python': datatypes.GregorianYearMonth.fromstring,
@@ -541,20 +419,20 @@ XSD_11_BUILTIN_TYPES = XSD_COMMON_BUILTIN_TYPES + (
     # --- Datetime derived types (XSD 1.1) ---
     {
         'name': XSD_DATE_TIME_STAMP,
-        'python_type': str,
+        'python_type': (datatypes.DateTimeStamp, str),
         'base_type': XSD_DATETIME,
         'to_python': datatypes.DateTime.fromstring,
         'facets': [etree_element(XSD_EXPLICIT_TIMEZONE, value='required')],
     },  # [-][Y*]YYYY-MM-DD[Thh:mm:ss] with required timezone
     {
         'name': XSD_DAY_TIME_DURATION,
-        'python_type': str,
+        'python_type': (datatypes.DayTimeDuration, str),
         'base_type': XSD_DURATION,
         'to_python': datatypes.DayTimeDuration.fromstring,
     },  # PnYnMnDTnHnMnS with month an year equal to 0
     {
         'name': XSD_YEAR_MONTH_DURATION,
-        'python_type': str,
+        'python_type': (datatypes.YearMonthDuration, str),
         'base_type': XSD_DURATION,
         'to_python': datatypes.YearMonthDuration.fromstring,
     },  # PnYnMnDTnHnMnS with day and time equals to 0
@@ -568,7 +446,10 @@ XSD_11_BUILTIN_TYPES = XSD_COMMON_BUILTIN_TYPES + (
 )
 
 
-def xsd_builtin_types_factory(meta_schema, xsd_types, atomic_builtin_class=None):
+def xsd_builtin_types_factory(
+        meta_schema: SchemaType,
+        xsd_types: Dict[str, Union[BaseXsdType, Tuple[ElementType, SchemaType]]],
+        atomic_builtin_class: Optional[Type[XsdAtomicBuiltin]] = None) -> None:
     """
     Builds the dictionary for XML Schema built-in types mapping.
     """
@@ -589,7 +470,7 @@ def xsd_builtin_types_factory(meta_schema, xsd_types, atomic_builtin_class=None)
 
     # xs:anySimpleType
     # Ref: https://www.w3.org/TR/xmlschema11-2/#builtin-stds
-    xsd_types[XSD_ANY_SIMPLE_TYPE] = XsdSimpleType(
+    xsd_any_simple_type = xsd_types[XSD_ANY_SIMPLE_TYPE] = XsdSimpleType(
         elem=etree_element(XSD_SIMPLE_TYPE, name=XSD_ANY_SIMPLE_TYPE),
         schema=meta_schema,
         parent=None,
@@ -598,46 +479,48 @@ def xsd_builtin_types_factory(meta_schema, xsd_types, atomic_builtin_class=None)
 
     # xs:anyAtomicType
     # Ref: https://www.w3.org/TR/xmlschema11-2/#builtin-stds
-    xsd_types[XSD_ANY_ATOMIC_TYPE] = meta_schema.BUILDERS.restriction_class(
+    xsd_types[XSD_ANY_ATOMIC_TYPE] = meta_schema.xsd_atomic_restriction_class(
         elem=etree_element(XSD_SIMPLE_TYPE, name=XSD_ANY_ATOMIC_TYPE),
         schema=meta_schema,
         parent=None,
         name=XSD_ANY_ATOMIC_TYPE,
-        base_type=xsd_types[XSD_ANY_SIMPLE_TYPE]
+        base_type=xsd_any_simple_type,
     )
 
     for item in builtin_types:
         item = item.copy()
-        name = item['name']
+        name: str = item['name']
         try:
-            elem, schema = xsd_types[name]
+            value = xsd_types[name]
         except KeyError:
             # If builtin type element is missing create a dummy element. Necessary for the
             # meta-schema XMLSchema.xsd of XSD 1.1, that not includes builtins declarations.
             elem = etree_element(XSD_SIMPLE_TYPE, name=name, id=name)
         else:
-            if schema is not meta_schema:
-                raise XMLSchemaValueError("loaded entry schema doesn't match meta_schema!")
+            if not isinstance(value, tuple):
+                continue
 
+            elem, schema = value
+            if schema is not meta_schema:
+                raise XMLSchemaValueError("loaded entry schema is not the meta-schema!")
+
+        base_type: Union[None, BaseXsdType, Tuple[ElementType, SchemaType]]
         if 'base_type' in item:
             base_type = item['base_type'] = xsd_types[item['base_type']]
         else:
             base_type = None
 
         facets = item.pop('facets', None)
-        builtin_type = atomic_builtin_class(elem, meta_schema, **item)
-        if isinstance(facets, (list, tuple)):
+        builtin_type: XsdAtomicBuiltin = atomic_builtin_class(elem, meta_schema, **item)
+        if facets:
             built_facets = builtin_type.facets
             for e in facets:
-                if is_etree_element(e):
-                    cls = facets_map[e.tag]
-                    built_facets[e.tag] = cls(e, meta_schema, builtin_type, base_type)
-                elif callable(e):
-                    if None in facets:
-                        raise XMLSchemaValueError("Almost one callable for facet group!!")
+                try:
+                    cls: Any = facets_map[e.tag]
+                except AttributeError:
                     built_facets[None] = e
                 else:
-                    raise XMLSchemaValueError("Wrong type for item %r" % e)
+                    built_facets[e.tag] = cls(e, meta_schema, builtin_type, base_type)
             builtin_type.facets = built_facets
 
         xsd_types[name] = builtin_type
