@@ -8,37 +8,80 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 """
-This module defines a mixin class for enabling XPath on schemas.
+This module defines a proxy class and a mixin class for enabling XPath on schemas.
 """
+import sys
 from abc import abstractmethod
-from collections.abc import Sequence
+from typing import cast, overload, Any, Dict, Iterator, List, Optional, \
+    Sequence, Set, TypeVar, Union
 import re
-import threading
 
 from elementpath import AttributeNode, TypedElement, XPath2Parser, \
-    XPathSchemaContext, AbstractSchemaProxy
+    XPathSchemaContext, AbstractSchemaProxy, protocols
 
-from .namespaces import XSD_NAMESPACE
 from .exceptions import XMLSchemaValueError, XMLSchemaTypeError
+from .names import XSD_NAMESPACE
+from .aliases import NamespacesType, SchemaType, BaseXsdType, XPathElementType
+from .helpers import get_qname, local_name, get_prefixed_qname
+
+if sys.version_info < (3, 8):
+    XMLSchemaProtocol = SchemaType
+    ElementProtocol = XPathElementType
+    XsdTypeProtocol = BaseXsdType
+else:
+    from typing import runtime_checkable, Protocol
+
+    XsdTypeProtocol = protocols.XsdTypeProtocol
+
+    class XMLSchemaProtocol(protocols.XMLSchemaProtocol, Protocol):
+        attributes: Dict[str, Any]
+
+    @runtime_checkable
+    class ElementProtocol(protocols.ElementProtocol, Protocol):
+        schema: XMLSchemaProtocol
+        attributes: Dict[str, Any]
+
 
 _REGEX_TAG_POSITION = re.compile(r'\b\[\d+]')
 
 
-def iter_schema_nodes(root, with_root=True, with_attributes=False):
+def iter_schema_nodes(root: Union[XMLSchemaProtocol, ElementProtocol],
+                      with_root: bool = True,
+                      with_attributes: bool = False) \
+        -> Iterator[Union[XMLSchemaProtocol, ElementProtocol, AttributeNode]]:
     """
     Iteration function for schema nodes. It doesn't yield text nodes,
     that are always `None` for schema elements, and detects visited
     element in order to skip already visited nodes.
 
-    :param root: schema or schema's element.
+    :param root: schema or schema element.
     :param with_root: if `True` yields initial element.
     :param with_attributes: if `True` yields also attribute nodes.
     """
-    def attribute_node(x):
+    def attribute_node(x: Any) -> AttributeNode:
         return AttributeNode(*x)
 
-    def _iter_schema_nodes(elem):
-        for child in elem:
+    if isinstance(root, TypedElement):
+        root = cast(ElementProtocol, root.elem)
+
+    nodes = {root}
+    if with_root:
+        yield root
+    if with_attributes:
+        yield from map(attribute_node, root.attributes.items())
+
+    iterators: List[Any] = []
+    children: Iterator[Any] = iter(root)
+
+    while True:
+        try:
+            child = next(children)
+        except StopIteration:
+            try:
+                children = iterators.pop()
+            except IndexError:
+                return
+        else:
             if child in nodes:
                 continue
             elif child.ref is not None:
@@ -49,23 +92,15 @@ def iter_schema_nodes(root, with_root=True, with_attributes=False):
                     yield child.ref
                     if with_attributes:
                         yield from map(attribute_node, child.attributes.items())
-                    yield from _iter_schema_nodes(child.ref)
+                    iterators.append(children)
+                    children = iter(child.ref)
             else:
                 nodes.add(child)
                 yield child
                 if with_attributes:
                     yield from map(attribute_node, child.attributes.items())
-                yield from _iter_schema_nodes(child)
-
-    if isinstance(root, TypedElement):
-        root = root.elem
-
-    nodes = {root}
-    if with_root:
-        yield root
-    if with_attributes:
-        yield from map(attribute_node, root.attributes.items())
-    yield from _iter_schema_nodes(root)
+                iterators.append(children)
+                children = iter(child)
 
 
 class XMLSchemaContext(XPathSchemaContext):
@@ -75,10 +110,14 @@ class XMLSchemaContext(XPathSchemaContext):
 
 class XMLSchemaProxy(AbstractSchemaProxy):
     """XPath schema proxy for the *xmlschema* library."""
-    def __init__(self, schema=None, base_element=None):
+    _schema: SchemaType  # type: ignore[assignment]
+
+    def __init__(self, schema: Optional[XMLSchemaProtocol] = None,
+                 base_element: Optional[ElementProtocol] = None) -> None:
+
         if schema is None:
             from xmlschema import XMLSchema
-            schema = XMLSchema.meta_schema
+            schema = getattr(XMLSchema, 'meta_schema', None)
         super(XMLSchemaProxy, self).__init__(schema, base_element)
 
         if base_element is not None:
@@ -88,59 +127,33 @@ class XMLSchemaProxy(AbstractSchemaProxy):
             except AttributeError:
                 raise XMLSchemaTypeError("%r is not an XsdElement" % base_element)
 
-    @property
-    def xsd_version(self):
-        return self._schema.XSD_VERSION
+    def bind_parser(self, parser: XPath2Parser) -> None:
+        parser.schema = self
+        parser.symbol_table = dict(parser.__class__.symbol_table)
 
-    def bind_parser(self, parser):
-        if parser.schema is not self:
-            parser.schema = self
+        with self._schema.lock:
+            if self._schema.xpath_tokens is None:
+                self._schema.xpath_tokens = {
+                    xsd_type.name: parser.schema_constructor(xsd_type.name)
+                    for xsd_type in self.iter_atomic_types() if xsd_type.name
+                }
 
-        if self._schema.xpath_tokens is None:
-            parser.symbol_table = parser.__class__.symbol_table.copy()
-            for xsd_type in self.iter_atomic_types():
-                parser.schema_constructor(xsd_type.name)
-            self._schema.xpath_tokens = parser.symbol_table
-        else:
-            parser.symbol_table = self._schema.xpath_tokens
-        parser.tokenizer = parser.create_tokenizer(parser.symbol_table)
+        parser.symbol_table.update(self._schema.xpath_tokens)
 
-    def get_context(self):
+    def get_context(self) -> XMLSchemaContext:
         return XMLSchemaContext(
-            root=self._schema,
-            namespaces=self._schema.namespaces,
+            root=self._schema,  # type: ignore[arg-type]
+            namespaces=dict(self._schema.namespaces),
             item=self._base_element
         )
 
-    def get_type(self, qname):
-        try:
-            return self._schema.maps.types[qname]
-        except KeyError:
-            return None
-
-    def get_attribute(self, qname):
-        try:
-            return self._schema.maps.attributes[qname]
-        except KeyError:
-            return None
-
-    def get_element(self, qname):
-        try:
-            return self._schema.maps.elements[qname]
-        except KeyError:
-            return None
-
-    def get_substitution_group(self, qname):
-        try:
-            return self._schema.maps.substitution_groups[qname]
-        except KeyError:
-            return None
-
-    def find(self, path, namespaces=None):
-        return self._schema.find(path, namespaces)
-
-    def is_instance(self, obj, type_qname):
+    def is_instance(self, obj: Any, type_qname: str) -> bool:
+        # FIXME: use elementpath.datatypes for checking atomic datatypes
         xsd_type = self._schema.maps.types[type_qname]
+        if isinstance(xsd_type, tuple):
+            from .validators import XMLSchemaNotBuiltError
+            raise XMLSchemaNotBuiltError(xsd_type[1], f"XSD type {type_qname} is not built")
+
         try:
             xsd_type.encode(obj)
         except ValueError:
@@ -148,84 +161,85 @@ class XMLSchemaProxy(AbstractSchemaProxy):
         else:
             return True
 
-    def cast_as(self, obj, type_qname):
+    def cast_as(self, obj: Any, type_qname: str) -> Any:
         xsd_type = self._schema.maps.types[type_qname]
+        if isinstance(xsd_type, tuple):
+            from .validators import XMLSchemaNotBuiltError
+            raise XMLSchemaNotBuiltError(xsd_type[1], f"XSD type {type_qname} is not built")
         return xsd_type.decode(obj)
 
-    def iter_atomic_types(self):
+    def iter_atomic_types(self) -> Iterator[XsdTypeProtocol]:
         for xsd_type in self._schema.maps.types.values():
-            if xsd_type.target_namespace != XSD_NAMESPACE and hasattr(xsd_type, 'primitive_type'):
-                yield xsd_type
+            if not isinstance(xsd_type, tuple) and \
+                    xsd_type.target_namespace != XSD_NAMESPACE and \
+                    hasattr(xsd_type, 'primitive_type'):
+                yield cast(XsdTypeProtocol, xsd_type)
 
-    def get_primitive_type(self, xsd_type):
-        return xsd_type.root_type
+    def get_primitive_type(self, xsd_type: XsdTypeProtocol) -> XsdTypeProtocol:
+        primitive_type = cast(BaseXsdType, xsd_type).root_type
+        return cast(XsdTypeProtocol, primitive_type)
 
 
-class ElementPathMixin(Sequence):
+E = TypeVar('E', bound='ElementPathMixin[Any]')
+
+
+class ElementPathMixin(Sequence[E]):
     """
-    Mixin abstract class for enabling ElementTree and XPath API on XSD components.
+    Mixin abstract class for enabling ElementTree and XPath 2.0 API on XSD components.
 
     :cvar text: the Element text, for compatibility with the ElementTree API.
     :cvar tail: the Element tail, for compatibility with the ElementTree API.
     """
-    text = None
-    tail = None
-    attributes = {}
-    namespaces = {}
+    text: Optional[str] = None
+    tail: Optional[str] = None
+    name: Optional[str] = None
+    attributes: Any = {}
+    namespaces: Any = {}
     xpath_default_namespace = ''
 
-    _xpath_parser = None  # Internal XPath 2.0 parser, instantiated at first use.
-
-    def __init__(self):
-        self._xpath_lock = threading.Lock()  # Lock for XPath operations
-
-    def __getstate__(self):
-        state = self.__dict__.copy()
-        state.pop('_xpath_lock', None)
-        state.pop('_xpath_parser', None)
-        state.pop('xpath_tokens', None)  # For schema objects
-        return state
-
-    def __setstate__(self, state):
-        self.__dict__.update(state)
-        self._xpath_lock = threading.Lock()
-
     @abstractmethod
-    def __iter__(self):
+    def __iter__(self) -> Iterator[E]:
         raise NotImplementedError
 
-    def __getitem__(self, i):
+    @overload
+    def __getitem__(self, i: int) -> E: ...
+
+    @overload
+    def __getitem__(self, s: slice) -> Sequence[E]: ...
+
+    def __getitem__(self, i: Union[int, slice]) -> Union[E, Sequence[E]]:
         try:
             return [e for e in self][i]
         except IndexError:
             raise IndexError('child index out of range')
 
-    def __reversed__(self):
+    def __reversed__(self) -> Iterator[E]:
         return reversed([e for e in self])
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len([e for e in self])
 
     @property
-    def tag(self):
+    def tag(self) -> str:
         """Alias of the *name* attribute. For compatibility with the ElementTree API."""
-        return getattr(self, 'name')
+        return self.name or ''
 
     @property
-    def attrib(self):
+    def attrib(self) -> Any:
         """Returns the Element attributes. For compatibility with the ElementTree API."""
         return self.attributes
 
-    def get(self, key, default=None):
+    def get(self, key: str, default: Any = None) -> Any:
         """Gets an Element attribute. For compatibility with the ElementTree API."""
         return self.attributes.get(key, default)
 
     @property
-    def xpath_proxy(self):
+    def xpath_proxy(self) -> XMLSchemaProxy:
         """Returns an XPath proxy instance bound with the schema."""
         raise NotImplementedError
 
-    def _get_xpath_namespaces(self, namespaces=None):
+    def _get_xpath_namespaces(self, namespaces: Optional[NamespacesType] = None) \
+            -> Dict[str, str]:
         """
         Returns a dictionary with namespaces for XPath selection.
 
@@ -238,35 +252,32 @@ class ElementPathMixin(Sequence):
         elif '' not in namespaces:
             namespaces[''] = self.xpath_default_namespace
 
-        xpath_namespaces = XPath2Parser.DEFAULT_NAMESPACES.copy()
+        xpath_namespaces: Dict[str, str] = XPath2Parser.DEFAULT_NAMESPACES.copy()
         xpath_namespaces.update(namespaces)
         return xpath_namespaces
 
-    def _xpath_parse(self, path, namespaces=None):
-        path = _REGEX_TAG_POSITION.sub('', path.strip())  # Strips tags positions from path
+    def is_matching(self, name: Optional[str], default_namespace: Optional[str] = None) -> bool:
+        if not name or name[0] == '{' or not default_namespace:
+            return self.name == name
+        else:
+            return self.name == '{%s}%s' % (default_namespace, name)
 
-        namespaces = self._get_xpath_namespaces(namespaces)
-        with self._xpath_lock:
-            parser = self._xpath_parser
-            if parser is None:
-                parser = XPath2Parser(namespaces, strict=False, schema=self.xpath_proxy)
-                self._xpath_parser = parser
-            else:
-                parser.namespaces = namespaces
-            return parser.parse(path)
-
-    def find(self, path, namespaces=None):
+    def find(self, path: str, namespaces: Optional[NamespacesType] = None) -> Optional[E]:
         """
         Finds the first XSD subelement matching the path.
 
         :param path: an XPath expression that considers the XSD component as the root element.
         :param namespaces: an optional mapping from namespace prefix to namespace URI.
-        :return: The first matching XSD subelement or ``None`` if there is not match.
+        :return: the first matching XSD subelement or ``None`` if there is no match.
         """
-        context = XMLSchemaContext(self)
-        return next(self._xpath_parse(path, namespaces).select_results(context), None)
+        path = _REGEX_TAG_POSITION.sub('', path.strip())  # Strips tags positions from path
+        namespaces = self._get_xpath_namespaces(namespaces)
+        parser = XPath2Parser(namespaces, strict=False)
+        context = XMLSchemaContext(self)  # type: ignore[arg-type]
 
-    def findall(self, path, namespaces=None):
+        return cast(Optional[E], next(parser.parse(path).select_results(context), None))
+
+    def findall(self, path: str, namespaces: Optional[NamespacesType] = None) -> List[E]:
         """
         Finds all XSD subelements matching the path.
 
@@ -275,10 +286,14 @@ class ElementPathMixin(Sequence):
         :return: a list containing all matching XSD subelements in document order, an empty \
         list is returned if there is no match.
         """
-        context = XMLSchemaContext(self)
-        return self._xpath_parse(path, namespaces).get_results(context)
+        path = _REGEX_TAG_POSITION.sub('', path.strip())  # Strips tags positions from path
+        namespaces = self._get_xpath_namespaces(namespaces)
+        parser = XPath2Parser(namespaces, strict=False)
+        context = XMLSchemaContext(self)  # type: ignore[arg-type]
 
-    def iterfind(self, path, namespaces=None):
+        return cast(List[E], parser.parse(path).get_results(context))
+
+    def iterfind(self, path: str, namespaces: Optional[NamespacesType] = None) -> Iterator[E]:
         """
         Creates and iterator for all XSD subelements matching the path.
 
@@ -286,17 +301,21 @@ class ElementPathMixin(Sequence):
         :param namespaces: is an optional mapping from namespace prefix to full name.
         :return: an iterable yielding all matching XSD subelements in document order.
         """
-        context = XMLSchemaContext(self)
-        return self._xpath_parse(path, namespaces).select_results(context)
+        path = _REGEX_TAG_POSITION.sub('', path.strip())  # Strips tags positions from path
+        namespaces = self._get_xpath_namespaces(namespaces)
+        parser = XPath2Parser(namespaces, strict=False)
+        context = XMLSchemaContext(self)  # type: ignore[arg-type]
 
-    def iter(self, tag=None):
+        return cast(Iterator[E], parser.parse(path).select_results(context))
+
+    def iter(self, tag: Optional[str] = None) -> Iterator[E]:
         """
         Creates an iterator for the XSD element and its subelements. If tag is not `None` or '*',
         only XSD elements whose matches tag are returned from the iterator. Local elements are
         expanded without repetitions. Element references are not expanded because the global
         elements are not descendants of other elements.
         """
-        def safe_iter(elem):
+        def safe_iter(elem: Any) -> Iterator[E]:
             if tag is None or elem.is_matching(tag):
                 yield elem
             for child in elem:
@@ -311,10 +330,10 @@ class ElementPathMixin(Sequence):
 
         if tag == '*':
             tag = None
-        local_elements = set()
+        local_elements: Set[E] = set()
         return safe_iter(self)
 
-    def iterchildren(self, tag=None):
+    def iterchildren(self, tag: Optional[str] = None) -> Iterator[E]:
         """
         Creates an iterator for the child elements of the XSD component. If *tag* is not `None`
         or '*', only XSD elements whose name matches tag are returned from the iterator.
@@ -324,3 +343,49 @@ class ElementPathMixin(Sequence):
         for child in self:
             if tag is None or child.is_matching(tag):
                 yield child
+
+
+class XPathElement(ElementPathMixin['XPathElement']):
+    """An element node for making XPath operations on schema types."""
+    name: str
+    parent = None
+
+    def __init__(self, name: str, xsd_type: BaseXsdType) -> None:
+        self.name = name
+        self.type = xsd_type
+        self.attributes = getattr(xsd_type, 'attributes', {})
+
+    def __iter__(self) -> Iterator['XPathElement']:
+        if not self.type.has_simple_content():
+            yield from self.type.content.iter_elements()  # type: ignore[union-attr]
+
+    @property
+    def xpath_proxy(self) -> XMLSchemaProxy:
+        return XMLSchemaProxy(
+            cast(XMLSchemaProtocol, self.schema),
+            cast(ElementProtocol, self)
+        )
+
+    @property
+    def schema(self) -> SchemaType:
+        return self.type.schema
+
+    @property
+    def target_namespace(self) -> str:
+        return self.type.schema.target_namespace
+
+    @property
+    def namespaces(self) -> NamespacesType:
+        return self.type.schema.namespaces
+
+    @property
+    def local_name(self) -> str:
+        return local_name(self.name)
+
+    @property
+    def qualified_name(self) -> str:
+        return get_qname(self.target_namespace, self.name)
+
+    @property
+    def prefixed_name(self) -> str:
+        return get_prefixed_qname(self.name, self.namespaces)
